/*
mesh.c
basic mesh manipulating routines
*/

#include <stdlib.h>
#include <math.h>

#include "triangles.h"

/* extern in triangles.c */
int error_exit(char *m1,char *m2);

float vdot(VECTOR v1,VECTOR v2){
   return (v1.x*v2.x + v1.y*v2.y + v1.z*v2.z);
}

void vcross(float x1, float y1, float z1, float x2, float y2, float z2, float *xout, float *yout, float *zout)
{
    *xout = (y1 * z2) - (z1 * y2);
    *yout = (z1 * x2) - (x1 * z2);
    *zout = (x1 * y2) - (y1 * x2);
}
float vlength(float x, float y, float z)
{
   return sqrt(x*x + y*y + z*z);
}

void vnormalize(float *x, float *y, float *z)
{
   float length = 1.0 / vlength(*x, *y, *z);

   *x *= length;
   *y *= length;
   *z *= length;
}

void  triangle_normal_V(VTX v1, VTX v2, VTX v3,VECTOR *tnorm){

   vcross(v2.x-v1.x, v2.y-v1.y, v2.z-v1.z,
          v3.x-v1.x, v3.y-v1.y, v3.z-v1.z,
          &(tnorm->x),&(tnorm->y),&(tnorm->z));
   vnormalize(&(tnorm->x),&(tnorm->y),&(tnorm->z));
}
/* requires already allocated VECTOR TNormals[number of triangles] */
int get_triangle_normals(TRIANGLE * TArray,VECTOR *TNormals, int TNum){
    TRIANGLE *T;
    VTX vA,vB,vC;
    int i;
    for(i=0;i<TNum;i++){
       T=&(TArray[i]);
       vA=T->vtxlist[T->v1];
       vB=T->vtxlist[T->v2];
       vC=T->vtxlist[T->v3];
       triangle_normal_V(vA, vB, vC,&(TNormals[i]) );
    }
    return 1;
}

/* requires already allocated VECTOR VNormals[number of vertices] */
int get_vertex_normals(TRIANGLE *TArray,int TNum, int VNum,VECTOR *VNormals,int AreaAvg){
    int i;
    int pA,pB,pC;
    TRIANGLE *T;
    VTX *VArray;
    VTX vA,vB,vC;
    VECTOR vN,nvN;
    for(i=0;i<VNum;i++){
         VNormals[i].x=0;VNormals[i].y=0;VNormals[i].z=0;
    }
    for (i=0;i<TNum;i++){

         T=&(TArray[i]);
         VArray=T->vtxlist;
         pA=T->v1; pB=T->v2; pC=T->v3;
         vA=VArray[pA];vB=VArray[pB];vC=VArray[pC];
         /* get vector normal to triangle i's surface */
         vcross(vB.x-vA.x,vB.y-vA.y,vB.z-vA.z, vC.x-vA.x,vC.y-vA.y,vC.z-vA.z, &(vN.x),&(vN.y),&(vN.z));
         nvN.x=vN.x; nvN.y=vN.y; nvN.z=vN.z;
         /* vN's length is proportional to triangle's area */
         if(!AreaAvg)
           vnormalize(&(nvN.x),&(nvN.y),&(nvN.z));
         if(!isnanf(nvN.x) && !isnanf(nvN.y) && !isnanf(nvN.z)){
            VNormals[pA].x += nvN.x; VNormals[pA].y += nvN.y; VNormals[pA].z += nvN.z;
            VNormals[pB].x += nvN.x; VNormals[pB].y += nvN.y; VNormals[pB].z += nvN.z;
            VNormals[pC].x += nvN.x; VNormals[pC].y += nvN.y; VNormals[pC].z += nvN.z;
         }

    }
    for(i=0;i<VNum;i++){
      if( !(VNormals[i].x==0 && VNormals[i].y && VNormals[i].z) ){
         vnormalize(&(VNormals[i].x),&(VNormals[i].y),&(VNormals[i].z));
      }
    }
    return 1;
}

void get_mesh_extent(MESH *m, float *max_x,float *max_y,float *max_z,float *min_x,float *min_y,float *min_z){
     int vi;
     VTX v=m->points[0];
     *max_x=v.x;
     *min_x=v.x;
     *max_y=v.y;
     *min_y=v.y;
     *max_z=v.z;
     *min_z=v.z;

     for (vi=1;vi<m->num_vertices;vi++){
           v=m->points[vi];
           if(v.x>*max_x)*max_x=v.x;
           if(v.y>*max_y)*max_y=v.y;
           if(v.z>*max_z)*max_z=v.z;
           if(v.x<*min_x)*min_x=v.x;
           if(v.y<*min_y)*min_y=v.y;
           if(v.z<*min_z)*min_z=v.z;
     }
}

/*****/

void init_shape(MESH *m,VECTOR v, VECTOR r, float dz, VECTOR dr,float scale);


MESH *new_mesh(int n_v,int n_f,int n_e,VECTOR v, VECTOR r, float dz, VECTOR dr,float scale){

      MESH *newmesh=(MESH *)malloc(sizeof(MESH));
      if(newmesh==NULL){
         return NULL;
      }
      newmesh->num_vertices=0;
      newmesh->num_faces=0;
      newmesh->num_edges=0;
      
      newmesh->num_vertices_allocated=n_v;
      newmesh->num_faces_allocated=n_f;
      newmesh->num_edges_allocated=n_e;

      newmesh->points=(VTX *)malloc(sizeof( VTX)*n_v);
      newmesh->faces=(TRIANGLE *)malloc(sizeof( TRIANGLE)*n_f);
      newmesh->edges=(EDGE *)malloc(sizeof(EDGE)*n_e);
      
      newmesh->output_points=(VTX *)malloc(sizeof( VTX)*n_v);
      newmesh->output_faces=(TRIANGLE *)malloc(sizeof( TRIANGLE)*n_f);


      
      if(   newmesh->points==NULL
         || newmesh->faces==NULL
         || newmesh->output_points==NULL
         || newmesh->output_faces==NULL
         || newmesh->edges==NULL)
      {
          error_exit("memory allocation error for new mesh",NULL);
      }

      init_shape(newmesh,v,r,dz,dr,scale);
      

      return newmesh;
}

void delete_mesh(MESH *m){
     if(m->points)
       free(m->points);
     if(m->faces)
       free(m->faces);
     if(m->output_points)
       free(m->output_points);
     if(m->output_faces)
       free(m->output_faces);
     if(m->edges)
       free(m->edges);
     free(m);
}
int new_vertex(MESH *m,VTX vtx,int f){
     if(m->num_vertices>=m->num_vertices_allocated-1){
        error_exit("Out of vertex space!",NULL);
     }
     m->points[m->num_vertices]=vtx;
     m->points[m->num_vertices].flags=f;
     m->num_vertices++;
     
     return (m->num_vertices-1);
}


int new_triangle(MESH *m,int v1, int v2, int v3,int oflag){
     int mynum=m->num_faces;
     if(m->num_faces>=m->num_faces_allocated-1){
        error_exit("Out of triangle space!",NULL);
     }

     m->faces[m->num_faces].vtxlist=m->points;
     m->faces[m->num_faces].v1=v1;
     m->faces[m->num_faces].v2=v2;
     m->faces[m->num_faces].v3=v3;
     m->faces[m->num_faces].oflag=oflag;
     m->num_faces++;
     return mynum;
}

void delete_triangle(MESH *m,int trinum){
     int ti;
     for(ti=trinum;ti<m->num_faces;ti++){
                   m->faces[ti]=m->faces[ti+1];
     }
     m->num_faces--;
}

void delete_vertex(MESH *m,int vtxnum){
     int f,v;
     int nf;
     TRIANGLE *T;
     f=0;
     nf=m->num_faces;
     while(f<m->num_faces){ /* first remove triangles containing vtxnum */
          T=&(m->faces[f]);

          if(T->v1==vtxnum||T->v2==vtxnum||T->v3==vtxnum){
               delete_triangle(m,f);
          }
          else
             f++;
     }
     for(f=0;f<m->num_faces;f++){
     /* then fix vertex numbers in triangles.
        Other lists of vertex numbers
        (like curVertex) need to be fixed, too. */
         T=&(m->faces[f]);
         if(T->v1 > vtxnum)
                  T->v1--;
         if(T->v2 > vtxnum)
                  T->v2--;
         if(T->v3 > vtxnum)
                  T->v3--;         
     }

     /* Finally, actually remove the vertex from the list */
     for (v=vtxnum;v<m->num_vertices;v++){
          m->points[v]=m->points[v+1];
     }

     m->num_vertices--;
}

int new_edge(MESH *m,int v1,int v2, int f){
     int mynum;
     if(m->num_edges>=m->num_edges_allocated-1){
        return -1;
     }
     mynum=m->num_edges;
     m->edges[mynum].v1=v1;
     m->edges[mynum].v2=v2;
     m->edges[mynum].t1=-1;
     m->edges[mynum].t2=-1;
     m->edges[mynum].f=f;
     m->edges[mynum].midindex=-1;
     m->num_edges++;
     return mynum;
}

float vtx2vtx_dist(MESH *m,int v1,int v2){
      float dx=m->points[v1].x-m->points[v2].x,
            dy=m->points[v1].y-m->points[v2].y,
            dz=m->points[v1].z-m->points[v2].z;
      return sqrt(dx*dx+dy*dy+dz*dz);
}

int add_mesh_to_mesh(MESH *dest,MESH *source){
      int vi,fi,vn;
      int v1,v2,v3;
      TRIANGLE T;
      VTX v;
      int n_v=source->num_vertices,
          n_f=source->num_faces;
      int prev_n_v=dest->num_vertices;

      if(dest->num_vertices+n_v>=dest->num_vertices_allocated-1){
          return -1;
      }
      if(dest->num_faces+n_f>=dest->num_faces_allocated-1){
          return -2;
      }
      if(source==NULL){
          return -3;
      }

      for(vi=0;vi<n_v;vi++){
          v=source->points[vi];
          vn=new_vertex(dest,v,0);
      }
      for(fi=0;fi<n_f;fi++){
          T=source->faces[fi];
          v1=T.v1;v2=T.v2;v3=T.v3;
          new_triangle(dest,v1+prev_n_v,v2+prev_n_v,v3+prev_n_v,0);
      }
      return n_v;
}

int triangle_has_vertex_p(MESH *m,int ti,int  vi){
    TRIANGLE T=m->faces[ti];
    return(T.v1==vi || T.v2==vi || T.v3==vi);
}

void reverse_triangle(TRIANGLE * T){
/* reverse the order of vertices */
      int v=T->v1;
      T->v1=T->v3;
      T->v3=v;
}

int find_triangle_in_mesh(MESH *m,int v1, int v2, int v3){
/* if found, returns triangle's number. if not, returns -1 */
     int f;
     TRIANGLE T;

     for(f=0;f<m->num_faces;f++){
           T=m->faces[f];
           if(T.v1==v1){
                if(T.v2==v2){
                   if(T.v3==v3){
                      return f;
                   }
                }
                else if(T.v2==v3){
                   if(T.v3==v2){
                      return f;
                   }
                }
           }
           else if(T.v1==v2){
                if(T.v2==v3){
                   if(T.v3==v1){
                      return f;
                   }
                }
                else if(T.v2==v1){
                   if(T.v3==v3){
                      return f;
                   }
                }

           }
           else if(T.v1==v3){
                if(T.v2==v2){
                   if(T.v3==v1){
                      return f;
                   }
                }
                else if(T.v2==v1){
                   if(T.v3==v2){
                      return f;
                   }
                }
           }
     }
     return -1; 
}
int find_edge_in_meshedges(MESH *m,int va,int vb, int *o){
/* if found, returns the edge's index; returns -1 if not found */
     int e;
     *o=0;
     for(e=0;e<m->num_edges;e++){
         if(m->edges[e].v1==va && m->edges[e].v2==vb){
              *o=1;
              return e; /* return index of found */
         }
         else if(m->edges[e].v1==vb && m->edges[e].v2==va){
              *o=-1;
              return e; /* return index of found */
         }
     }
     return -1; /* not found */
}

int find_edge_in_triangle(TRIANGLE * T,  int va,int vb){
      int v1=T->v1,v2=T->v2,v3=T->v3;
      if((va==v1 && vb==v2)||(va==v2 && vb==v3)||(va==v3 && vb==v1))
         return 1; /* found as va,vb */
      else if((vb==v1 && va==v2)||(vb==v2 && va==v3)||(vb==v3 && va==v1))
         return -1; /* found as vb,va */
      return 0; /* not found */
}


/* A special routine to run when edges are needed.
   Mostly, the program doesn't care about edges */
int make_edge_list(MESH *m){
/* returns: 1 (ok), 0 (singular), -1 (out of edges!) */
      int f;
      TRIANGLE *T;
      int v1,v2,v3;
      int o,e;
      int ok=1;
      m->num_edges=0;
      for(f=0;f<m->num_faces;f++){
          T=&(m->faces[f]);
          v1=T->v1;v2=T->v2;v3=T->v3;
          if((e=find_edge_in_meshedges(m,v1,v2,&o))<0){
               e=new_edge(m,v1,v2,0);
               if(e<0) return -1;
               m->edges[e].t1=f;
          }
          else{
               if(m->edges[e].t2!=-1) ok=0;
               m->edges[e].t2=f;// if more than two triangles share this
                                // edge, this is nearly worthless
          }
          if((e=find_edge_in_meshedges(m,v2,v3,&o))<0){
               e=new_edge(m,v2,v3,0);
               if(e<0) return -1;
               m->edges[e].t1=f;
          }
          else{
               if(m->edges[e].t2!=-1) ok=0;
               m->edges[e].t2=f;
          }

          if((e=find_edge_in_meshedges(m,v3,v1,&o))<0){
               e=new_edge(m,v3,v1,0);
               if(e<0) return -1;
               m->edges[e].t1=f;
          }
          else{
               if(m->edges[e].t2!=-1) ok=0;
               m->edges[e].t2=f;
          }

      }
      return ok;
}



void clear_all_orientflags(MESH *m){
     int f;
     for (f=0;f<m->num_faces;f++){
          m->faces[f].oflag=0;
     }
}

int recurse_orient(MESH *m,TRIANGLE *T){
     int v1=T->v1,v2=T->v2,v3=T->v3;
     int i,o;
     TRIANGLE *T1;
     // for each of T's edges, look for triangles
     // sharing that edge.
     for(i=0;i<m->num_faces;i++){
         T1=&(m->faces[i]);
         if(T1!=T){
            o=find_edge_in_triangle(T1, v1,v2);
            if(o && !(T1->oflag &1) ){
            // if T1 shares an edge and it hasn't already been oriented
                 if(o>0)
                     reverse_triangle(T1);
                 T1->oflag|=1;
                 recurse_orient(m,T1);
            }
         }
     }
     for(i=0;i<m->num_faces;i++){
         T1=&(m->faces[i]);
         if(T1!=T){
            o=find_edge_in_triangle(T1, v2,v3);
            if(o && !(T1->oflag&1)){
                 if(o>0)
                     reverse_triangle(T1);
                 T1->oflag|=1;
                 recurse_orient(m,T1);
            }
         }
     }
     for(i=0;i<m->num_faces;i++){
         T1=&(m->faces[i]);
         if(T1!=T){
            o=find_edge_in_triangle(T1, v3,v1);
            if(o && !(T1->oflag&1)){
                 if(o>0)
                     reverse_triangle(T1);
                 T1->oflag|=1;
                 recurse_orient(m,T1);
            }
         }
     }
     return 0;
}
int try_orient(MESH *m){
    int t;
    TRIANGLE *T;
    clear_all_orientflags(m);
    for(t=0;t<m->num_faces;t++){
          T=&(m->faces[t]);
          if(!T->oflag&1){
             T->oflag|=1;
             recurse_orient(m,T);
          }
    }
    return 0;
}

int try_orient_triangle_w_neighbors(MESH *m,TRIANGLE *T){
     int f,o=-1;
     TRIANGLE *Tn=NULL,*Tt;
     int va,vb;
     
     va=T->v1;vb=T->v2;
     for (f=0;f<m->num_faces;f++){
         Tt=&(m->faces[f]);
         if(Tt!=T){
           o=find_edge_in_triangle(Tt,  va, vb);
           if(o && Tt->oflag){
              Tn=Tt;
              break;
           }
         }
     }
     if(Tn){
         if(o>0)
            reverse_triangle(T);
         T->oflag=1;
         return 1;//success
     }
     va=T->v2;vb=T->v3;
     for (f=0;f<m->num_faces;f++){
         Tt=&(m->faces[f]);
         if(Tt!=T){
           o=find_edge_in_triangle(Tt,  va, vb);
           if(o && Tt->oflag){
              Tn=Tt;
              break;
           }
         }
     }

     if(Tn){
         if(o>0)
            reverse_triangle(T);
         T->oflag=1;
         return 1;//success
     }
     va=T->v3;vb=T->v1;
     for (f=0;f<m->num_faces;f++){
         Tt=&(m->faces[f]);
         if(Tt!=T){
           o=find_edge_in_triangle(Tt,  va, vb);
           if(o && Tt->oflag){
              Tn=Tt;
              break;
           }
         }
     }

     if(Tn){
         if(o>0)
            reverse_triangle(T);
         T->oflag=1;
         return 1;//success
     }
     
     return 0;//failure
}

int new_triangle_o(MESH *m,int v1, int v2, int v3){
     TRIANGLE * newT;
     int mynum=new_triangle(m,v1,v2,v3,0);
     newT=&(m->faces[m->num_faces-1]);
     if(m->num_faces==1)
        newT->oflag=1;
     else
        try_orient_triangle_w_neighbors(m,newT);
     return mynum;
}


int divide_triangle_mesh(MESH *m){
     TRIANGLE *T1;
     int ti;
     float v1x,v1y,v1z,v2x,v2y,v2z,v3x,v3y,v3z;
     int v1,v2,v3,e;
     int newvtxnum1,newvtxnum2,newvtxnum3;
     VTX newvtx1,newvtx2,newvtx3;
     int numV=m->num_vertices,
         numT=m->num_faces,
         numE;
     int newnumE,
         newnumV,
         newnumT=4*numT;

     EDGE * NewE;
     
     int okedges=make_edge_list(m);
     if(okedges==-1)
        return -3;
     numE=m->num_edges;
     
     newnumE=2*numE+numT*3;
     newnumV=numV+numE;
     if(newnumE >= m->num_edges_allocated-1){
          return -3;
     }
   
     if(newnumV >= m->num_vertices_allocated-1){
          return -2;
     }
   
     if(newnumT >= m->num_faces_allocated-1){
          return -1;
     }

     NewE=(EDGE *)malloc(sizeof(EDGE)*newnumE);
     if(NewE ==NULL){
          return -4;
     }
     for(e=0;e<newnumE;e++){
         NewE[e].midindex=-1;
     }
     numE=0;
     for(ti=0;ti<numT;ti++){
        T1=&(m->faces[ti]);
        v1=T1->v1;v2=T1->v2;v3=T1->v3;
        v1x=m->points[v1].x;
        v1y=m->points[v1].y;
        v1z=m->points[v1].z;
        v2x=m->points[v2].x;
        v2y=m->points[v2].y;
        v2z=m->points[v2].z;
        v3x=m->points[v3].x;
        v3y=m->points[v3].y;
        v3z=m->points[v3].z;
        newvtxnum1=-1;newvtxnum2=-1;newvtxnum3=-1;
        for(e=0;e<numE;e++){

            if( (NewE[e].v1==v1&&NewE[e].v2==v2) || (NewE[e].v1==v2&&NewE[e].v2==v1)){
                newvtxnum1=NewE[e].midindex;
            }
            /***/
            if( (NewE[e].v1==v2&&NewE[e].v2==v3) || (NewE[e].v1==v3&&NewE[e].v2==v2)){
                newvtxnum2=NewE[e].midindex;
            }
            if( (NewE[e].v1==v3&&NewE[e].v2==v1) || (NewE[e].v1==v1&&NewE[e].v2==v3)){
                newvtxnum3=NewE[e].midindex;
            }

        }
        if(newvtxnum1==-1){
                newvtx1.x=(v1x+v2x)/2.0;
                newvtx1.y=(v1y+v2y)/2.0;
                newvtx1.z=(v1z+v2z)/2.0;
                newvtxnum1=new_vertex(m,newvtx1,0);
                NewE[numE].v1=v1;NewE[numE].v2=v2;
                NewE[numE].midindex=newvtxnum1;
                numE++;
        }
        if(newvtxnum2==-1){
                newvtx2.x=(v2x+v3x)/2.0;
                newvtx2.y=(v2y+v3y)/2.0;
                newvtx2.z=(v2z+v3z)/2.0;

                newvtxnum2=new_vertex(m,newvtx2,0);
                NewE[numE].v1=v2;NewE[numE].v2=v3;
                NewE[numE].midindex=newvtxnum2;
                numE++;

        }

        if(newvtxnum3==-1){
                newvtx3.x=(v3x+v1x)/2.0;
                newvtx3.y=(v3y+v1y)/2.0;
                newvtx3.z=(v3z+v1z)/2.0;
                newvtxnum3=new_vertex(m,newvtx3,0);
                NewE[numE].v1=v3;NewE[numE].v2=v1;
                NewE[numE].midindex=newvtxnum3;
                numE++;

        }
        T1->v1=newvtxnum1;T1->v2=newvtxnum2;T1->v3=newvtxnum3;
        new_triangle(m,v1,newvtxnum1,newvtxnum3,2);
        new_triangle(m,newvtxnum1,v2,newvtxnum2,2);
        new_triangle(m,newvtxnum2,v3,newvtxnum3,2);

     }
     free (NewE);
     return 0;
}


void insert_vertex(MESH *m,int va,int vb,VTX vtx){
    int fi,o;
    int v1=-1,v2=-1,v3=-1, // initialize to supress warning
        v4;
    TRIANGLE *T;
    v4=new_vertex(m,vtx,0);
    for(fi=0;fi<m->num_faces;fi++){
       T=&(m->faces[fi]);
       if( (o=find_edge_in_triangle(T,va,vb)) ){

           if(T->v1==va){
              if(o>0){
                 v1=va;v2=vb;v3=T->v3;
              }
              else{
                 v1=vb;v2=va;v3=T->v2;
              }
           }
           else if(T->v2==va){
              if(o>0){
                 v1=va;v2=vb;v3=T->v1;
              }
              else{
                 v1=vb;v2=va;v3=T->v3;
              }
           }
           else if(T->v3==va){
              if(o>0){
                 v1=va;v2=vb;v3=T->v2;
              }
              else{
                 v1=vb;v2=va;v3=T->v1;
              }
           }
           T->v1=v1;T->v2=v4;T->v3=v3;
           new_triangle(m,v2,v3,v4,0);
           
       }
    }

}

void init_shape(MESH *m,VECTOR v, VECTOR r, float dz, VECTOR dr,float scale)
{

   m->shape.x = v.x;
   m->shape.y = v.y;
   m->shape.z = v.z;
   m->shape.rx = r.x;
   m->shape.ry = r.y;
   m->shape.rz = r.z;
   m->shape.dz =  dz;
   m->shape.drx = dr.x;
   m->shape.dry = dr.y;
   m->shape.drz = dr.z;
   m->shape.scale=scale;
}
void copy_shape(MESH *D,MESH * S){
      D->shape.x=S->shape.x;
      D->shape.y=S->shape.y;
      D->shape.z=S->shape.z;
      D->shape.dz=S->shape.dz;
      D->shape.rx = S->shape.rx;
      D->shape.ry =S->shape.ry;
      D->shape.rz =S->shape.rz;
      D->shape.drx=S->shape.drx;
      D->shape.dry=S->shape.dry;
      D->shape.drz=S->shape.drz;     
}

void animate_shape(MESH *m)
{
      m->shape.z += m->shape.dz;

      if ((m->shape.z > (1024)) ||
	  (m->shape.z < (192)))
	 m->shape.dz = -m->shape.dz;

      m->shape.rx += m->shape.drx;
      m->shape.ry += m->shape.dry;
      m->shape.rz += m->shape.drz;
}
void ani_rotate(MESH *m,VECTOR dr)
{
      m->shape.rx += dr.x;
      m->shape.ry += dr.y;
      m->shape.rz += dr.z;
}
void ani_translate(MESH *m,VECTOR dr)
{
      m->shape.x += dr.x;
      m->shape.y += dr.y;
      if( m->shape.z > 10 || (m->shape.z <= 10 && dr.z>0) )
         m->shape.z += dr.z;      
}
void ani_scale(MESH *m,float scale)
{
      m->shape.scale *= scale;
}

void remove_duplicate_triangles(MESH *m){
    int i,j;
    int deleteme=0;
    int v1,v2,v3;
    TRIANGLE *T1,*T2;
    for(i=0;i<m->num_faces;i++){
         T1=&(m->faces[i]);
         v1=T1->v1;v2=T1->v2;v3=T1->v3;
         for(j=i+1;j<m->num_faces;j++){
            T2=&(m->faces[j]);

            if(v1==T2->v1){
                 if(v2==T2->v2){
                     if(v3==T2->v3)
                        deleteme=1;
                 }
                 else if(v2==T2->v3){
                     if(v3==T2->v2)
                        deleteme=1;
                 }
            }
            else if(v1==T2->v2){
                 if(v2==T2->v1){
                     if(v3==T2->v3)
                        deleteme=1;
                 }
                 else if(v2==T2->v3){
                     if(v3==T2->v1)
                        deleteme=1;
                 }
            }
            else if(v1==T2->v3){
                 if(v2==T2->v1){
                     if(v3==T2->v2)
                        deleteme=1;
                 }
                 else if(v2==T2->v2){
                     if(v3==T2->v1)
                        deleteme=1;
                 }
            }
            if(deleteme){
                 delete_triangle(m,j);
                 deleteme=0;
                 j--;
            }
         }
    }
    for(i=0;i<m->num_faces;i++){
         T1=&(m->faces[i]);
         v1=T1->v1;v2=T1->v2;v3=T1->v3;
         if(v1==v2 || v1==v3 || v2==v3){
            delete_triangle(m,i);
            i--;
         }
    }
}
int vertex_used_p(MESH *m,int vi){
/* is vertex vi in any triangle of mesh m? */
    int vtxused=0;
    int ti;
    TRIANGLE T;

    for(ti=0;ti<m->num_faces;ti++){
              T=m->faces[ti];
              if(T.v1==vi||T.v2==vi||T.v3==vi){
                  vtxused=1;
                  break;
              }
    }

    return vtxused;
}

int meld_vertices(MESH *m, int i1,int i2){
    TRIANGLE *T;
    int keep,meld,f;
    if(i1==i2) return -1;

    keep=i1;meld=i2;
    for(f=0;f<m->num_faces;f++){
           T =&(m->faces[f]);
           if(T->v1==meld)T->v1=keep;
           if(T->v2==meld)T->v2=keep;
           if(T->v3==meld)T->v3=keep;
    }
    delete_vertex(m,meld);

    remove_duplicate_triangles(m);
    return keep;
}


float DistPointToSegment2d(float px, float py, float X1, float Y1, float X2, float Y2){
   float dx,dy,t;

   dx = X2 - X1;
   dy = Y2 - Y1;
   if(dx==0 && dy==0){
        /* It's a point, not a line segment. */
        dx = px - X1;
        dy = py - Y1;
        return sqrt(dx * dx + dy * dy);
    }
    t = (dx * (px - X1) + dy * (py - Y1)) / (dx * dx + dy * dy);
    if(t < 0){
        dx = px - X1;
        dy = py - Y1;
    }
    else if (t > 1){
        dx = px - X2;
        dy = py - Y2;
    }
    else{
        X2 = X1 + t * dx;
        Y2 = Y1 + t * dy;
        dx = px - X2;
        dy = py - Y2;
    }
    return sqrt(dx * dx + dy * dy);
}

void mesh_CG(MESH *m,VTX *cg){
     int i;
     int n_v=m->num_vertices;
     cg->x=0;cg->y=0;cg->z=0;
     for (i=0;i<n_v;i++){
             cg->x+=m->points[i].x;
             cg->y+=m->points[i].y;
             cg->z+=m->points[i].z;
     }
     cg->x/=(float)n_v;
     cg->y/=(float)n_v;
     cg->z/=(float)n_v;

}

int *connected=NULL;
void recurse_connected(MESH *m,int vtx){
/* find all vertices connected to vtx by some edge-path */
       int ti;

       TRIANGLE T;
       connected[vtx]=1;


       for (ti=0; ti < m->num_faces ; ti++) {
           T=m->faces[ti];
           if(connected[T.v1]||connected[T.v2]||connected[T.v3]){
                if(vtx!=T.v1 && !connected[T.v1]){
                  connected[T.v1]=1;
                  recurse_connected(m,T.v1);
                }
                if(vtx!=T.v2 && !connected[T.v2]){
                  connected[T.v2]=1;
                  recurse_connected(m,T.v2);
                }
                if(vtx!=T.v3 && !connected[T.v3]){
                  connected[T.v3]=1;
                  recurse_connected(m,T.v3);
                }
           }
       }
}

/* returns new mesh containing the connected component
   of MESH *m that vertex number vtx is in. The calling
   routine is responsible for freeing (delete_mesh())
   this new mesh*/
MESH * copy_of_connected_component(MESH *m,int vtx){

       MESH *cmesh=NULL;
       int n_v=0,
           n_f=0;
       int num_connected=0;
       int vi,ti;
       int v1,v2,v3;
       VTX v;
       TRIANGLE T;

       connected=malloc(sizeof(int)*m->num_vertices);

       if(connected==NULL)
           return NULL;
       for(vi=0;vi<m->num_vertices;vi++){
           connected[vi]=0;
       }
       recurse_connected(m,vtx);
       for(vi=0;vi<m->num_vertices;vi++){
          if(connected[vi]){
             num_connected++;
             m->points[vi].flags|=2;
          }
       }
       
       cmesh=(MESH *)malloc(sizeof(MESH));
       if(cmesh==NULL){
          free(connected); connected=NULL;
          error_exit("memory allocation error for mesh copy",NULL);
       }

       cmesh->output_points=NULL;
       cmesh->output_faces=NULL;
       cmesh->edges=NULL;

       cmesh->num_vertices=0;
       cmesh->num_faces=0;
       
       n_v=m->num_vertices;
       n_f=m->num_faces;
       cmesh->num_vertices_allocated=n_v+1;
       cmesh->num_faces_allocated=n_f+1;

       cmesh->points=(VTX *)malloc(sizeof( VTX)*(n_v+1));
       cmesh->faces=(TRIANGLE *)malloc(sizeof( TRIANGLE)*(n_f+1));

      
       if(cmesh->points==NULL || (n_f && cmesh->faces==NULL))
       {
          free(connected); connected=NULL;
          error_exit("memory allocation error for mesh copy",NULL);
       }

       for(vi=0;vi<m->num_vertices;vi++){
           if(connected[vi]){
              v=m->points[vi];
              new_vertex(cmesh,v,0);
           }
       }
       for(ti=0;ti<m->num_faces;ti++){
           T=m->faces[ti];
           if(connected[T.v1]){

               v1=0;
               v2=0;
               v3=0;
               vi=0;
               while(v1<num_connected){
                    if(T.v1==vi){
                       break;
                    }
                    if(connected[vi]){
                       v1++;
                    }
                    vi++;
               }
               vi=0;
               while(v1<num_connected){
                    if(T.v2==vi){
                       break;
                    }
                    if(connected[vi]){
                       v2++;
                    }
                    vi++;
               }
               vi=0;
               while(v3<num_connected){
                    if(T.v3==vi){
                       break;
                    }
                    if(connected[vi]){
                       v3++;
                    }
                    vi++;
               }
               if(v1<num_connected&&v2<num_connected&&v3<num_connected){
                  new_triangle(cmesh,v1,v2,v3,0);
               }

           }
       }


       free(connected); connected=NULL;
       return cmesh;
}

