/* poly2tri.c
   'Entry point' to a polygon triangulation algorithm
   (in construct.c, monotone.c, and misc.c) by A. Narkhede
   and D. Manocha from Graphics Gems V:
      An implementation of a fast polygon triangulation algorithm
      based on the paper "A simple and fast incremental randomized
      algorithm for computing trapezoidal decompositions and for
      triangulating polygons" by Raimund Seidel.
      
   poly2tri.c. also contains the 'latheing' routine
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "allegro.h"

#include "triangles.h"

#include "triangulate.h"

#define MAXCYCLEPOINTS SEGSIZE
#define MAXCYCLES SEGSIZE


extern int VIEWS_WIDTH;
extern int VIEWS_HEIGHT;
int POLYVIEW_WIDTH=0;
int POLYVIEW_HEIGHT=0;
void set_polygon_dialog_sizes(int w, int h);
/* external in triangles.c */
int error_exit(char *message1,char *message2);
int do_snapval_dialog();
extern MESH *mesh;
extern int num_selected;
int new_vertex(MESH *m,VTX vtx,int f);
int new_triangle(MESH *m,int v1, int v2, int v3,int oflag);
void add_to_selected (int vtxnum, int removeiffound);
/***/

typedef struct{
   float x,y;
}pVTX;

typedef struct{
   pVTX * polypoints;
   int num_points;
   int cycle;
}pCycle;

pCycle *polycycles;
int num_cycles;
int curCycle;

static int initialise_triangulation(n)
     int n;
{
  register int i;

  for (i = 1; i <= n; i++)
    seg[i].is_inserted = FALSE;

  generate_random_ordering(n);
  
  return 0;
}

int rescale_output=1;

int triangulate_poly(){
  int nv;
  VTX v;
  int n, nmonpoly, genus;
  int op[SEGSIZE][3], i, ntriangles;
  int ccount;
  int ncontours, npoints, first, last;
  float rescale=rescale_output?((float)VIEWS_WIDTH/(float)POLYVIEW_WIDTH):1.0;
  ncontours=num_cycles;

  if (ncontours <= 0)
    return -1;
  
  /* For every contour, read in all the points for the contour. The
   * outer-most contour is read in first (points specified in
   * anti-clockwise order). Next, the inner contours are input in
   * clockwise order */

  ccount = 0;
  i = 1;
  
  while (ccount < ncontours){
      int j;

      npoints=polycycles[ccount].num_points;
      first = i;
      last = first + npoints - 1;
      for (j = 0; j < npoints; j++, i++){
          seg[i].v0.x= polycycles[ccount].polypoints[j].x;
          seg[i].v0.y= polycycles[ccount].polypoints[j].y;
          if (i == last)
	    {
	      seg[i].next = first;
	      seg[i].prev = i-1;
	      seg[i-1].v1 = seg[i].v0;
	    }
	  else if (i == first)
	    {
	      seg[i].next = i+1;
	      seg[i].prev = last;
	      seg[last].v1 = seg[i].v0;
	    }
	  else
	    {
	      seg[i].prev = i-1;
	      seg[i].next = i+1;
	      seg[i-1].v1 = seg[i].v0;
	    }
	  
	  seg[i].is_inserted = FALSE;
      }
      ccount++;
  }

  genus = ncontours - 1;
  n= i-1;

  num_selected=0;
  nv=mesh->num_vertices;
  for(i=1;i<=n;i++){
        v.x=seg[i].v0.x * rescale;
        v.y=seg[i].v0.y * rescale;
        v.z=0.0;
        new_vertex(mesh,v,0);
        add_to_selected(mesh->num_vertices-1,0);
  }
  
  initialise_triangulation(n);
  construct_trapezoids(n);
  nmonpoly = monotonate_trapezoids(n);
  ntriangles = triangulate_monotone_polygons(n, nmonpoly, op);

  for (i = 0; i < ntriangles; i++){
        new_triangle(mesh,nv+op[i][0]-1,nv+op[i][1]-1,nv+op[i][2]-1,0);
  }
  return 0;
}

int num_lathe_segments=12;
int lathe_poly(){
  int nv;
  VTX v;
  int vn=0,firstvn=0; /* initialize to suppress warning */
  int v1,v2,v3,v4;
  int s;
  int i, ccount,npoints;
  double a,cosa,sina,r;
  int totalnewvertices=0;
  float rescale=rescale_output?((float)VIEWS_WIDTH/(float)POLYVIEW_WIDTH):1.0;
  for(ccount=0;ccount<num_cycles;ccount++){
      totalnewvertices+=polycycles[ccount].num_points*num_lathe_segments;
      
  }
  if(mesh->num_vertices+totalnewvertices>=mesh->num_vertices_allocated-1){
       alert("Too many vertices to lathe",NULL,NULL,"OK",NULL,13,0);
       return -1;
  }
  if(mesh->num_faces+2*totalnewvertices>=mesh->num_faces_allocated-1){
       alert("Too many triangles to lathe",NULL,NULL,"OK",NULL,13,0);
       return -1;
  }

  if(num_lathe_segments<3)
      return -1;
  nv=mesh->num_vertices;

  for(ccount=0;ccount<num_cycles;ccount++){

     npoints=polycycles[ccount].num_points;
     for(s=0;s<num_lathe_segments;s++){
        a=(double)s*2*PI/(double)num_lathe_segments;
        cosa=cos(a);sina=sin(a);
        
        for(i=0;i<npoints;i++){
           r=polycycles[ccount].polypoints[i].x*rescale;
           v.x=r*cosa;
           v.y=polycycles[ccount].polypoints[i].y*rescale;
           v.z=r*sina;
           vn=new_vertex(mesh,v,0);
           if(s>0){
              if(i>0){
                  v3=vn-1;
                  v4=vn;
                  v1=v3-npoints;
                  v2=v4-npoints;
                  new_triangle(mesh,v1,v2,v3,0);
                  new_triangle(mesh,v3,v2,v4,0);
                  if(s==num_lathe_segments-1){
                      v1=v3;
                      v2=v4;
                      v4=firstvn+i;
                      v3=v4-1;
                      new_triangle(mesh,v1,v2,v3,0);
                      new_triangle(mesh,v3,v2,v4,0);
                  }
              }              
           }
           else if(i==0){
              firstvn=vn;
           }
        }
        if(polycycles[ccount].cycle){
          if(s>0){
            v3=vn;
            v4=vn+1-npoints;
            v1=v3-npoints;
            v2=v4-npoints;
            new_triangle(mesh,v1,v2,v3,0);
            new_triangle(mesh,v3,v2,v4,0);
            if(s==num_lathe_segments-1){
                      v1=v3;
                      v2=v4;
                      v4=firstvn;
                      v3=firstvn+npoints-1;
                      new_triangle(mesh,v1,v2,v3,0);
                      new_triangle(mesh,v3,v2,v4,0);
            }
          }
        }
     }
         
  }
    
  num_selected=0;
  for(i=nv;i<mesh->num_vertices;i++){
     add_to_selected(i,0);
  }


  
  return 0;
}

int init_poly_pipe(){
  int nv;
  VTX v;
  int vn=0,firstvn=0; /* initialize to suppress warning */
  int v1,v2,v3,v4;
  int s;
  int i, ccount,npoints;
  int totalnewvertices=0;
  float rescale=rescale_output?((float)VIEWS_WIDTH/(float)POLYVIEW_WIDTH):1.0;
  for(ccount=0;ccount<num_cycles;ccount++){
      totalnewvertices+=polycycles[ccount].num_points*2;
      
  }
  if(mesh->num_vertices+totalnewvertices>=mesh->num_vertices_allocated-1){
       alert("Too many vertices to extrude",NULL,NULL,"OK",NULL,13,0);
       return -1;
  }
  if(mesh->num_faces+2*totalnewvertices>=mesh->num_faces_allocated-1){
       alert("Too many triangles to extrude",NULL,NULL,"OK",NULL,13,0);
       return -1;
  }

  nv=mesh->num_vertices;
  num_selected=0;
  for(ccount=0;ccount<num_cycles;ccount++){

     npoints=polycycles[ccount].num_points;
     s=0;
     while(s<2){
        nv=mesh->num_vertices;
        for(i=0;i<npoints;i++){
   
           v.x=polycycles[ccount].polypoints[i].x*rescale;;
           v.y=polycycles[ccount].polypoints[i].y*rescale;
           v.z=0;
           vn=new_vertex(mesh,v,0);
           if(s>0){
              if(i>0){
                  v3=vn-1;
                  v4=vn;
                  v1=v3-npoints;
                  v2=v4-npoints;
                  new_triangle(mesh,v1,v2,v3,0);
                  new_triangle(mesh,v3,v2,v4,0);
                  if(s==num_lathe_segments-1){
                      v1=v3;
                      v2=v4;
                      v4=firstvn+i;
                      v3=v4-1;
                      new_triangle(mesh,v1,v2,v3,0);
                      new_triangle(mesh,v3,v2,v4,0);
                  }
              }              
           }
           else if(i==0){
              firstvn=vn;
           }
        }
        if(polycycles[ccount].cycle){
          if(s>0){
            v3=vn;
            v4=vn+1-npoints;
            v1=v3-npoints;
            v2=v4-npoints;
            new_triangle(mesh,v1,v2,v3,0);
            new_triangle(mesh,v3,v2,v4,0);
            if(s==num_lathe_segments-1){
                      v1=v3;
                      v2=v4;
                      v4=firstvn;
                      v3=firstvn+npoints-1;
                      new_triangle(mesh,v1,v2,v3,0);
                      new_triangle(mesh,v3,v2,v4,0);
            }
          }
        }
        if(s==0){

           for(i=nv;i<mesh->num_vertices;i++){
                add_to_selected(i,0);
           }

        }
        s++;

     }
  }
  
  return 0;
}


int SegSegIntersect(pVTX A, pVTX B, pVTX C, pVTX D);
void init_polycycles(){
    int c;
    for(c=0;c<MAXCYCLES;c++){
        polycycles[c].polypoints=NULL;
        polycycles[c].num_points=0;
        polycycles[c].cycle=0;        
    }
    num_cycles=0;
    curCycle=0;
}


int new_cycle(){
    int ret;
    pVTX *pvtx;
    if(num_cycles>=MAXCYCLES)
        return -1;
    pvtx=malloc(sizeof(pVTX)*MAXCYCLEPOINTS);
    if(pvtx==NULL)
          error_exit("memory allocation error in new_cycle",NULL);
    polycycles[num_cycles].polypoints=pvtx;
    polycycles[num_cycles].cycle=0;
    ret=num_cycles;
    num_cycles++;
    return ret;
}

void kill_cycle(int pc_index){
    int c;
    pCycle *pc=&(polycycles[pc_index]);
    if(pc->polypoints){
        free(pc->polypoints);
        pc->polypoints=NULL;
    }
    for(c=pc_index;c<num_cycles;c++){
        polycycles[c]=polycycles[c+1];
    }
    num_cycles--;

}

void delete_all_cycles(){
    int c;
    pCycle *pc;
    for(c=0;c<MAXCYCLES;c++){
        pc=&(polycycles[c]);
        if(pc->polypoints){
           free(pc->polypoints);
           pc->polypoints=NULL;
        }
    }
    num_cycles=0;
}

int get_maxy_cyclepoint(pCycle* pc){
   int i;
   int maxy_index=0;
   float maxy;
   if(pc->num_points==0)
       return -1;
   maxy=pc->polypoints[0].y;
   for(i=1;i<pc->num_points;i++){

        if(pc->polypoints[i].y>maxy){
              maxy=pc->polypoints[i].y;
              maxy_index=i;
        }
   }
   return maxy_index;

}
/* ======= Crossings algorithm ============================================ */
/* adapted from Graphics Gems IV, Eric Haines

 * Shoot a test ray along +X axis.  The strategy, from MacMartin, is to
 * compare vertex Y values to the testing point's Y and quickly discard
 * edges which are entirely to one side of the test ray.
 */

int CrossingsTest( pCycle *pcycle, float px, float py )
{
    int nv=pcycle->num_points;
    int	i,j, yflag0, yflag1, inside_flag, xflag0 ;

    pVTX vtx0,vtx1;
    
    vtx0 = pcycle->polypoints[nv-1] ;
    /* get test bit for above/below X axis */
    yflag0 = ( vtx0.y >= py ) ;
    vtx1 = pcycle->polypoints[0] ;

    inside_flag = 0 ;
    i=0;
    for ( j = nv+1 ; --j ; ) {
	yflag1 = ( vtx1.y >= py ) ;
	/* check if endpoints straddle (are on opposite sides) of X axis
	 * (i.e. the Y's differ); if so, +X ray could intersect this edge.*/
	if ( yflag0 != yflag1 ) {
	    xflag0 = ( vtx0.x >= px ) ;
	    /* check if endpoints are on same side of the Y axis (i.e. X's
	     * are the same); if so, it's easy to test if edge hits or misses.*/
	    if ( xflag0 == ( vtx1.x >= px ) ) {

		/* if edge's X values both right of the point, must hit */
		if ( xflag0 ) 
                     inside_flag = !inside_flag ;
	    } 
            else {
		/* compute intersection of pgon segment with +X ray, note
		 * if >= point's X; if so, the ray hits it.
		 */
		if ( (vtx1.x - (vtx1.y-py)*
		     ( vtx0.x-vtx1.x)/(vtx0.y-vtx1.y)) >= px ) {
		    inside_flag = !inside_flag ;
		}
	    }
	}
	/* move to next pair of vertices, retaining info as possible */
	yflag0 = yflag1 ;
	vtx0 = vtx1 ;
        i++;
	vtx1 = pcycle->polypoints[i] ;
    }
    return( inside_flag ) ;
}

int clockwise(pCycle *pc){
 int n,maxy_point,next,mid,prev;
 pVTX vtx0,vtx1,vtx2;
 float z=0;
 
 n=pc->num_points;
 maxy_point=get_maxy_cyclepoint(pc);
 mid=maxy_point;
 while (z==0){
     next=(mid+1)%n;
     prev=(mid>0)?mid-1:n-1;
     vtx1=pc->polypoints[mid];
     vtx0=pc->polypoints[prev];
     vtx2=pc->polypoints[next];
     z  = (vtx1.x - vtx0.x) * (vtx2.y - vtx1.y)-(vtx1.y - vtx0.y) * (vtx2.x - vtx1.x);
     if(z<0)
        return -1;
     else if(z>0)
        return +1;
     mid=(mid+1)%n;
 }
 return 0;
}


int new_cyclepoint(pCycle *pc, pVTX vtx)
{
     if(pc->num_points>=MAXCYCLEPOINTS){
        return -1;
     }
     pc->polypoints[pc->num_points]=vtx;

     pc->num_points++;
     
     return (pc->num_points-1);
}

void reverse_cycle(pCycle *pc){
     int i;
     pVTX hold;
     for (i=0;i<pc->num_points/2;i++){
         hold=pc->polypoints[i];
         pc->polypoints[i]=pc->polypoints[pc->num_points-i-1];
         pc->polypoints[pc->num_points-i-1]=hold;
     }
}

BITMAP *poly1bmp=NULL;
BITMAP *poly2bmp=NULL;
int drawing;
#define VERTEXCOLOR 1
void draw_cycle(pCycle c){
   int i;
   float x1,y1,x2,y2;
   int ix1,iy1,ix2,iy2;
   int lcolor=255;

   if(c.cycle){
      if(clockwise(&c)==-1)
          lcolor=127;
   }
   for(i=0;i<c.num_points;i++){
      x1=c.polypoints[i].x;
      y1=c.polypoints[i].y;
      ix1=(int)x1; iy1=(int)y1;
      circle(poly2bmp,ix1+ POLYVIEW_WIDTH/2,iy1+ POLYVIEW_HEIGHT/2,2,VERTEXCOLOR);
      if(i<c.num_points-1){
          x2=c.polypoints[i+1].x;
          y2=c.polypoints[i+1].y;
          ix2=(int)x2; iy2=(int)y2;
          line(poly2bmp,ix1+ POLYVIEW_WIDTH/2,iy1+ POLYVIEW_HEIGHT/2,ix2+ POLYVIEW_WIDTH/2,iy2+ POLYVIEW_HEIGHT/2,lcolor);
      }
      else if(c.cycle){
          ix2=(int)c.polypoints[0].x;
          iy2=(int)c.polypoints[0].y;
          line(poly2bmp,ix1+ POLYVIEW_WIDTH/2,iy1+ POLYVIEW_HEIGHT/2,ix2+ POLYVIEW_WIDTH/2,iy2+ POLYVIEW_HEIGHT/2,lcolor);
      }
   }
}

void draw_poly(){
    int c;
    int cw;
    clear(poly2bmp);
    vline(poly2bmp,POLYVIEW_WIDTH/2,0,POLYVIEW_HEIGHT,150);
    hline(poly2bmp,0,POLYVIEW_HEIGHT/2,POLYVIEW_WIDTH,150);
    for(c=0;c<num_cycles;c++){
         if(polycycles[c].cycle){
            cw=clockwise(&(polycycles[c]));
            if(c>0 && cw==1)
               reverse_cycle(&(polycycles[c]));
            if(c==0 && cw==-1)
               reverse_cycle(&(polycycles[c]));
         }
         draw_cycle(polycycles[c]);
    }
}

char polymousetxt[256]="";
int polyview_proc(int msg, DIALOG *d, int c);
int donepoly_button_proc(int msg, DIALOG *d, int c);
int lathepoly_button_proc(int msg, DIALOG *d, int c);
int pipepoly_button_proc(int msg, DIALOG *d, int c);
int cancelpoly_button_proc(int msg, DIALOG *d, int c);
int clearall_button_proc(int msg, DIALOG *d, int c);
int clearcycle_button_proc(int msg, DIALOG *d, int c);

#define PBG 200
#define C(x) (x-'a'+1)

DIALOG poly_dialog[] =
{
   // (dialog proc)         (x)     (y)  (w)          (h)            (fg)     (bg)      (key)     (flags)  (d1)    (d2) (dp)

   { d_box_proc,            0, 0, 0,0,     4,      180,        0,       0,      0,       0, NULL ,NULL,NULL},

   { polyview_proc,         2, 2, 0, 0,        255,        0,        0,       0,      0,       0, NULL, NULL,NULL},
   { clearall_button_proc,  0,30, 80, 20,                   0,      PBG,      'a',  D_EXIT,      0,       0, "clear All", NULL,NULL},
   { clearcycle_button_proc,0,60, 100, 20,                  0,      PBG,      'c',  D_EXIT,      0,       0, "clear Cycle", NULL,NULL},

   { d_textbox_proc,
                           0,      110,    100,  24,     0,  180,   0,      0,          0,    0,  polymousetxt ,NULL,NULL},
   
   { d_check_proc,          0, 0, 150, 16,0,      180,      's',  D_SELECTED,  0,       0, "reScale output", NULL,NULL},
      
   { donepoly_button_proc,  0, 0,100, 20,  0,      PBG,      't',  D_EXIT,      0,       0, "Triangulate", NULL,NULL},
   { lathepoly_button_proc, 0, 0, 90, 20,  0,      PBG,      'l',  D_EXIT,      0,       0, "Lathe", NULL,NULL},
   { pipepoly_button_proc,  0, 0, 90, 20,  0,      PBG,      'l',  D_EXIT,      0,       0, "Pipe", NULL,NULL},
   { cancelpoly_button_proc,0, 0, 60, 20,  0,      PBG,       27,  D_EXIT,      0,       0, "Cancel", NULL,NULL},
   
   { NULL,                   0,             0,            0,    0,    0,        0,    0,    0,      0,      0, NULL ,NULL,NULL}
};

#define POLYMAINLAYOUT 0
#define POLYVIEW   1
#define POLYCLEARALL 2
#define POLYCLEARCYCLE 3
#define POLYMOUSETEXT 4
#define POLYRESCALECHECK 5
#define POLYDONE   6
#define LATHEPOLY  7
#define PIPEPOLY   8
#define POLYCANCEL 9


int donepoly_button_proc(int msg, DIALOG *d, int c){
    int bret=d_button_proc(msg,d,c);
    if(bret==D_CLOSE){

    }

    return bret;
}

char lathe_segments_string[12];
DIALOG lathe_dialog[] =
{
   // (dialog proc)         (x)     (y)  (w)          (h)            (fg)     (bg)      (key)     (flags)  (d1)    (d2) (dp)

   { d_box_proc,             20,      20, 150,         130,            9,      180,        0,         0,      0,       0, NULL ,NULL,NULL},
   { d_text_proc,            30,      40,   0,           0,            0,      180,        0,         0,      0,       0, "Lathe segments:",NULL,NULL},
   { d_edit_proc,            30,      60,  50,          20,            0,      255,        0,         0,      5,       0, lathe_segments_string,NULL,NULL},
   { d_button_proc,          30,     120,  60,          20,            0,      200,       13,    D_EXIT,      0,       0, "OK", NULL,NULL},
   { d_button_proc,         100,     120,  60,          20,            0,      200,       27,    D_EXIT,      0,       0, "Cancel", NULL,NULL},
   
   { NULL,                   0,             0,            0,    0,    0,        0,    0,    0,      0,      0, NULL ,NULL,NULL}
};

#define LATHESECTIONSEDIT 2
#define LATHEDONE 3
#define LATHECANCEL 4
int do_lathe_dialog(){
    int popupret;
    int newlathesections;
    sprintf(lathe_segments_string,"%d", num_lathe_segments);

    popupret=popup_dialog(lathe_dialog,3);
    if(popupret==LATHEDONE){
        newlathesections=atoi(lathe_segments_string);
        if(newlathesections>2){
           num_lathe_segments=newlathesections;
           return 1;
        }
        
    }
    return 0;
    
}

int lathepoly_button_proc(int msg, DIALOG *d, int c){
    int bret=d_button_proc(msg,d,c);
    if(bret==D_CLOSE){
        if(!do_lathe_dialog()){
            return D_O_K;
        }
    }

    return bret;
}
int pipepoly_button_proc(int msg, DIALOG *d, int c){
    int bret=d_button_proc(msg,d,c);
    if(bret==D_CLOSE){

    }

    return bret;
}

int cancelpoly_button_proc(int msg, DIALOG *d, int c){
    int bret=d_button_proc(msg,d,c);
    if(bret==D_CLOSE){

    }
    return bret;
}

int clearall_button_proc(int msg, DIALOG *d, int c UNUSED){
    int bret=d_button_proc(msg,d,c);
    if(bret==D_CLOSE){
        delete_all_cycles();
        init_polycycles();
        curCycle=new_cycle();
        show_mouse(NULL);
        SEND_MESSAGE(&(poly_dialog[POLYVIEW]),MSG_DRAW,0);
        show_mouse(screen);
        return D_O_K;
    }
    return bret;
}
int clearcycle_button_proc(int msg, DIALOG *d, int c UNUSED){
    int bret=d_button_proc(msg,d,c);
    if(bret==D_CLOSE){
        kill_cycle(curCycle);
        if(curCycle==0)
          curCycle=new_cycle();
        else
          curCycle--;
        show_mouse(NULL);
        SEND_MESSAGE(&(poly_dialog[POLYVIEW]),MSG_DRAW,0);
        show_mouse(screen);
        return D_O_K;
    }
    return bret;
}

void update_poly_mouse_numbers(int mx,int my){
    char mousextxt[50];
    char mouseytxt[50];


    DIALOG *dm=&(poly_dialog[POLYMOUSETEXT]);
    sprintf(mousextxt,"x %+5.2f",(float)mx);
    sprintf(mouseytxt,"y %+5.2f",(float)-my);

    sprintf(polymousetxt,"%s\n%s",mousextxt,mouseytxt);
    SEND_MESSAGE(dm,MSG_DRAW,0);

}

int lastnew_x=0,lastnew_y=0;

int polyview_proc(int msg, DIALOG *d, int c UNUSED){

   pVTX pv,A,B,C,D;
   pCycle *pc;
   pCycle pci;
      
   int vmx,vmy,mx,my,xmickeys,ymickeys;
   static int lastmx=0,lastmy=0;
   int ct;


   int i,j;
   if (msg==MSG_DRAW){

       draw_poly();
       blit(poly2bmp, poly1bmp, 0,0,0,0,POLYVIEW_WIDTH,POLYVIEW_HEIGHT);
       d_bitmap_proc(MSG_DRAW,d,0);
       return D_O_K;
       
   }
   else if(msg==MSG_IDLE){
       mx=mouse_x;my=mouse_y;
       if( mx >= d->x && mx < d->x+d->w && my >= d->y && my < d->y+d->h){
           if(mx!=lastmx||my!=lastmy){
              lastmx=mx;lastmy=my;
              vmx = mx - (d->x+POLYVIEW_WIDTH/2);
              vmy = my - (d->y+POLYVIEW_HEIGHT/2);
              update_poly_mouse_numbers(vmx,vmy);
           }

       }
       return D_O_K;
   }

   else if(msg==MSG_CLICK){
        pc=&(polycycles[curCycle]);
        mx=mouse_x;my=mouse_y;
        get_mouse_mickeys(&xmickeys, &ymickeys);

        vmx = mx-(d->x+POLYVIEW_WIDTH/2);
        vmy = my-(d->y+POLYVIEW_HEIGHT/2);
        if(key[KEY_LSHIFT]||key[KEY_RSHIFT]){
            if( abs(lastnew_x-vmx)>abs(lastnew_y-vmy) ){
                  vmy=lastnew_y;
            }
            else if( abs(lastnew_x-vmx)<abs(lastnew_y-vmy) ){
                  vmx=lastnew_x;
            }
        }
        if(mouse_b&1){
               if(curCycle>0){
                  /* make sure point *is* inside polycycle[0],*/
                  if(!CrossingsTest( &(polycycles[0]) , vmx,  vmy )){
                      return D_O_K;
                  }
                  /* and point is *not* inside any other finished polycycle,... */
                  for(i=1;i<curCycle;i++){
                        if(CrossingsTest( &(polycycles[i]) , vmx,  vmy ))
                             return D_O_K;
                  }
               }
               /* ...and that the new edge (if) does not cross
                  any other edges */
               if(pc->num_points>0 && !pc->cycle){
                  A=pc->polypoints[pc->num_points-1];
                  B.x=vmx;B.y=vmy;
                  /* check intersections w/ other cycles */
                  for(i=0;i<curCycle;i++){
                      pci=polycycles[i];
                      for(j=0;j<pci.num_points;j++){
   
                         C=pci.polypoints[j];
                         D=pci.polypoints[(j+1)%pci.num_points];
                         
                         if(SegSegIntersect(A,B,C,D))
                             return D_O_K;
                      }
                  }
                  /* check intersections with curCycle */
                  if(pc->num_points>2 ){
                      for(j=0;j<pc->num_points-2;j++){
   
                         C=pc->polypoints[j];
                         D=pc->polypoints[(j+1)];
                         
                         if(SegSegIntersect(A,B,C,D))
                             return D_O_K;
                      }

                  }
               }

               if(pc->cycle){
                  /* The current cycle is finished,
                     and the new vertex will be part of the next cycle,
                     so make sure it is properly inside/outside curCycle */
                  ct=CrossingsTest( &(polycycles[curCycle]) , vmx,  vmy );
                  if( (ct&&curCycle>0) || (!ct && curCycle==0))
                      return D_O_K;
                  curCycle=new_cycle();
                  pc=&(polycycles[curCycle]);
               }
               if(pc->num_points){
                  /* and make sure it is not coincident with other vertices */
                  for(i=0;i<pc->num_points;i++){
                       if(vmx==pc->polypoints[i].x && vmy==pc->polypoints[i].y){
                          return D_O_K;
                       }
                  }
               }
               pv.x=vmx;pv.y=vmy;
               new_cyclepoint(pc,pv);
               lastnew_x=vmx;lastnew_y=vmy;

               show_mouse(NULL);
               SEND_MESSAGE(d,MSG_DRAW,0);
               show_mouse(screen);
        }
        else if(mouse_b&2||mouse_b&4){
        /* finish the current cycle.*/
               if(pc->num_points<3){
                  /* if curCycle doesn't even make a triangle,
                     forget it. */
                  kill_cycle(curCycle);
                  if(curCycle==0)
                     curCycle=new_cycle();
                  else
                     curCycle--;
               }
               else{
                    /* make sure the closing segment (last->first)
                       is ok. */
                    A=pc->polypoints[pc->num_points-1];
                    B=pc->polypoints[0];
                    /* check intersections w/ other cycles */
                    for(i=0;i<curCycle;i++){
                        pci=polycycles[i];
                        for(j=0;j<pci.num_points;j++){
     
                           C=pci.polypoints[j];
                           D=pci.polypoints[(j+1)%pci.num_points];
                           
                           if(SegSegIntersect(A,B,C,D))
                               return D_O_K;
                        }
                    }
                    /* check intersections with curCycle */
                    if(pc->num_points>2){
                        for(j=1;j<pc->num_points-2;j++){
     
                           C=pc->polypoints[j];
                           D=pc->polypoints[(j+1)];
                           
                           if(SegSegIntersect(A,B,C,D))
                               return D_O_K;
                        }
  
                    }

                    pc->cycle=1;
               }
               show_mouse(NULL);
               SEND_MESSAGE(d,MSG_DRAW,0);
               show_mouse(screen);
        }
        return D_O_K;
        
   }
   else
      return D_O_K;
}

int do_poly_dialog(){
    int pret,ret=0;
    int i;

    if(POLYVIEW_WIDTH==0||POLYVIEW_HEIGHT==0)
       set_polygon_dialog_sizes((3*VIEWS_WIDTH/2),(3*VIEWS_HEIGHT/2));


    polycycles=malloc(sizeof(pCycle)*MAXCYCLES);
    if(polycycles==NULL)
        error_exit("polygon cycles memory allocation error",NULL);
    init_polycycles();
    curCycle=new_cycle();

    poly1bmp = create_bitmap(POLYVIEW_WIDTH, POLYVIEW_HEIGHT);
    poly2bmp = create_bitmap(POLYVIEW_WIDTH, POLYVIEW_HEIGHT);
    if(poly1bmp==NULL||poly2bmp==NULL){
       error_exit("polygon bitmap memory allocation error",NULL);
    }
    
    poly_dialog[POLYVIEW].dp=poly1bmp;
    pret=popup_dialog(poly_dialog,-1);
    
    rescale_output=poly_dialog[POLYRESCALECHECK].flags&D_SELECTED;
    if(pret==POLYDONE){
         for(i=0;i<num_cycles;i++){
             /* kill any cycles that aren't completed. */
             if(polycycles[i].cycle==0)
                 kill_cycle(i);
         }
         if(polycycles[0].num_points>2){
         /* if there are at least 3 points in the outer cycle
            go ahead and triangulate */
            triangulate_poly();
            ret= 1;
         }
    }
    else if(pret==LATHEPOLY){
         for(i=1;i<num_cycles;i++){
             /* ? kill any inner cycles that aren't completed. */
             if(polycycles[i].cycle==0){
                 //kill_cycle(i);
             }
         }
         if(polycycles[0].num_points>1){

             lathe_poly();
             ret=1;
         }
    }
    else if(pret==PIPEPOLY){
         for(i=1;i<num_cycles;i++){
             /* ? kill any inner cycles that aren't completed. */
             if(polycycles[i].cycle==0){
                 //kill_cycle(i);
             }
         }
         if(polycycles[0].num_points>1){

             init_poly_pipe();
             ret=1;
         }
    }

   /* clean up before leaving */
   for(i=0;i<num_cycles;i++){
       kill_cycle(i);
   }

   destroy_bitmap(poly1bmp);poly1bmp=NULL;
   destroy_bitmap(poly2bmp);poly2bmp=NULL;
   free(polycycles);
   return ret;
}



int SegSegIntersect(pVTX A, pVTX B, pVTX C, pVTX D){
   float r,s;
   float denom;
   float num1,num2;
   float minABy,maxABy,minCDy,maxCDy,minABx,maxABx,minCDx,maxCDx;
/*
    Let A,B,C,D be 2-space position vectors.  Then the directed line
    segments AB & CD are given by:

        AB=A+r(B-A), r in [0,1]
        CD=C+s(D-C), s in [0,1]

    If AB & CD intersect, then

        A+r(B-A)=C+s(D-C), or

        XA+r(XB-XA)=XC+s(XD-XC)
        YA+r(YB-YA)=YC+s(YD-YC)  for some r,s in [0,1]

    Solving the above for r and s yields

            (YA-YC)(XD-XC)-(XA-XC)(YD-YC)
        r = -----------------------------  (eqn 1)
            (XB-XA)(YD-YC)-(YB-YA)(XD-XC)
            
            (YA-YC)(XB-XA)-(XA-XC)(YB-YA)
        s = -----------------------------  (eqn 2)
            (XB-XA)(YD-YC)-(YB-YA)(XD-XC)
*/
     denom=(B.x-A.x)*(D.y-C.y)-(B.y-A.y)*(D.x-C.x);
/*
If the denominator in eqn 1 is zero, AB & CD are parallel
If the numerator in eqn 1 is also zero, AB & CD are coincident
*/
     num1=(A.y-C.y)*(D.x-C.x)-(A.x-C.x)*(D.y-C.y);
     num2=(A.y-C.y)*(B.x-A.x)-(A.x-C.x)*(B.y-A.y);
     if(denom==0){
         if(num1==0){
              minABy=A.y<B.y?A.y:B.y;
              maxABy=A.y>B.y?A.y:B.y;
              minCDy=C.y<D.y?C.y:D.y;
              maxCDy=C.y>D.y?C.y:D.y;
              if(minABy>maxCDy||minCDy>maxABy)
                   return 0;
              minABx=A.x<B.x?A.x:B.x;
              maxABx=A.x>B.x?A.x:B.x;
              minCDx=C.x<D.x?C.x:D.x;
              maxCDx=C.x>D.x?C.x:D.x;
              if(minABx>maxCDx||minCDx>maxABx)
                   return 0;

              return 1;
         }
         return 0;
     }

     r=num1/denom;
     s=num2/denom;
     if(r<=1 && r>=0 && s<=1 && s>=0)
        return 1;
     return 0;

}


void set_polygon_dialog_sizes(int pview_width, int pview_height){
     POLYVIEW_WIDTH= pview_width;
     POLYVIEW_HEIGHT= pview_height;

     poly_dialog[POLYMAINLAYOUT].x=0; poly_dialog[POLYMAINLAYOUT].y=0;
     poly_dialog[POLYMAINLAYOUT].w=POLYVIEW_WIDTH+154;
     poly_dialog[POLYMAINLAYOUT].h=POLYVIEW_HEIGHT+4;
     poly_dialog[POLYVIEW].x=2; poly_dialog[POLYVIEW].y=2;
     poly_dialog[POLYVIEW].w=POLYVIEW_WIDTH;poly_dialog[POLYVIEW].h=POLYVIEW_HEIGHT;
     poly_dialog[POLYCLEARALL].x=POLYVIEW_WIDTH+10;
     poly_dialog[POLYCLEARCYCLE].x=POLYVIEW_WIDTH+10;
     poly_dialog[POLYMOUSETEXT].x=POLYVIEW_WIDTH+10;
     poly_dialog[POLYRESCALECHECK].x=POLYVIEW_WIDTH+10; poly_dialog[POLYRESCALECHECK].y=POLYVIEW_HEIGHT-180;
     poly_dialog[POLYDONE].x=POLYVIEW_WIDTH+10;poly_dialog[POLYDONE].y=POLYVIEW_HEIGHT-120;
     poly_dialog[LATHEPOLY].x=POLYVIEW_WIDTH+10;poly_dialog[LATHEPOLY].y=POLYVIEW_HEIGHT-90;
     poly_dialog[PIPEPOLY].x=POLYVIEW_WIDTH+10;poly_dialog[PIPEPOLY].y=POLYVIEW_HEIGHT-60;
     poly_dialog[POLYCANCEL].x=POLYVIEW_WIDTH+80;poly_dialog[POLYCANCEL].y=POLYVIEW_HEIGHT-30;

     
}

