
/* reduce.c
 *
 *  An adaptation of Stan Melax's article
 *  "A Simple, Fast, and Efficient Polygon Reduction Algorithm"
 *  in Game Developer magazine, November 1998
 *
 *  If you want to use a progressive mesh reduction algorithm,
 *  I recommend using the original source from Stan Melax instead
 *  of this file.
 */
#include <stdlib.h>
#include <math.h>
#include "triangles.h"
#include "idx_list.h"

int error_exit(char *message1,char *message2);
/* external in mesh.c */
float vtx2vtx_dist(MESH *m,int u,int v);
int triangle_has_vertex_p(MESH *m,int ti,int  vi);
float vdot(VECTOR v1,VECTOR v2);
int get_triangle_normals(TRIANGLE * TArray,VECTOR *TNormals, int TNum);
void delete_vertex(MESH *m,int vtxnum);
int meld_vertices(MESH *m, int i1,int i2);
/*****/

void init_list(INDEX_LIST *L);
INDEX_LIST * newlist();
void kill_list(INDEX_LIST *L);
void destroy_list(INDEX_LIST *L);
void add_index_to_list(INDEX_LIST *L,int index);
void remove_index_from_list(INDEX_LIST *L,int index);

int * boundary=NULL;
int * collapse=NULL;       /* collapse[i] is the index of which vertex
                              vertex_i should collapse into.*/
float * objdist=NULL;      /* objdist[i] is cost of collapsing vertex_i
                              to collapse[i] */
INDEX_LIST *neighbor=NULL; /* neighbor [i] is a list of vertices
                              connected to vertex_i by triangle edges. */
INDEX_LIST * face=NULL;    /* face[i] is a list of triangles which
                              have vertex_i as a corner */


INDEX_LIST * init_neighbor_list(MESH *m){
   int i;
   int n_v=m->num_vertices;
   if(neighbor)
       free(neighbor);
   neighbor=malloc(sizeof(INDEX_LIST)*n_v);
   if(neighbor==NULL){
       return NULL;
   }

   for(i=0;i<n_v;i++){
       neighbor[i].length=0;
       neighbor[i].first=NULL;
       neighbor[i].last=NULL;
   }
   return neighbor;
}

int get_neighbors(MESH *m,INDEX_LIST *N,int v){

     int n=0;
     TRIANGLE *T;
     INDEX_LIST_ENTRY *curface=face[v].first;
     /* for each triangle vertex v is in ... */     
     while(curface!=NULL){
         T=&(m->faces[curface->index]);
         /* ... add the other vertices to v's neighbor list */
         
         if(v==T->v1){
              add_index_to_list(N,T->v2);
              n++;
              add_index_to_list(N,T->v3);
              n++;

         }
         else if(v==T->v2){
              add_index_to_list(N,T->v3);
              n++;
              add_index_to_list(N,T->v1);
              n++;

         }
         else if(v==T->v3){
              add_index_to_list(N,T->v1);
              n++;
              add_index_to_list(N,T->v2);
              n++;
         }
         
         curface=curface->next;

     }
     return n;
}
int fill_neighbor(MESH *m){
     int i;
     INDEX_LIST *L;
     if(!init_neighbor_list(m)) {
         return 0;
     }
     for(i=0;i<m->num_vertices;i++){
           L=&(neighbor[i]);
           get_neighbors(m,L,i);           
     }
     return 1;
}
void destroy_neighbor(MESH *m){
   int i;
   for(i=0;i<m->num_vertices;i++){
       kill_list(&(neighbor[i]));
   }
   free(neighbor);
   neighbor=NULL;
}
INDEX_LIST * init_face_list(MESH *m){
   int i;
   int n_v=m->num_vertices;
   if(face)
       free(face);
   face=malloc(sizeof(INDEX_LIST)*n_v);
   if(face==NULL){
       return NULL;
   }
   for(i=0;i<n_v;i++){
       face[i].length=0;
       face[i].first=NULL;
       face[i].last=NULL;
   }
   return face;
}
int fill_face(MESH *m){
   int ti;
   TRIANGLE *T;
   
   if(!init_face_list(m)) {
         return 0;
   }

   for(ti=0;ti<m->num_faces;ti++){
         T=&(m->faces[ti]);
         add_index_to_list(&(face[T->v1]),ti);
         add_index_to_list(&(face[T->v2]),ti);
         add_index_to_list(&(face[T->v3]),ti);

   }
   return 1;

}
void destroy_face(MESH *m){
   int i;
   for(i=0;i<m->num_vertices;i++){
       kill_list(&(face[i]));
   }
   free(face);
   face=NULL;
}

VECTOR * TNormals;
float min(float a,float b){
      if(a<b)
        return a;
      return b;
}
float max(float a,float b){
      if(a>b)
        return a;
      return b;
}

float ComputeEdgeCollapseCost(MESH *m,int u,int v) {
	/* if we collapse edge uv by moving u to v then how
	   much different will the model change, i.e. how much "error".
	   The method of determining cost was designed in order
	   to exploit small and coplanar regions for
	   effective polygon reduction. */
        float mincurv,dotprod;
	float edgelength = vtx2vtx_dist(m,u,v);
	float curvature=0;
        INDEX_LIST_ENTRY *uface;
        INDEX_LIST_ENTRY *uvside;

	/* find the "side" triangles that are on the edge uv */
	INDEX_LIST sides; init_list(&sides);
        uface=face[u].first;
        while(uface!=NULL){
            if(triangle_has_vertex_p(m,uface->index,v)){
                 add_index_to_list(&sides,uface->index);
            }
            
            uface=uface->next;
        }
	/* use the triangle facing most away from the sides
	   to determine our curvature term */
        uface=face[u].first;
        while(uface!=NULL){
             mincurv=1;
             uvside=sides.first;
             while(uvside!=NULL){
                 dotprod=vdot(TNormals[uface->index],TNormals[uvside->index]);
                 mincurv = min(mincurv,(1-dotprod)/2.0f);

                 uvside=uvside->next;
             }
             curvature = max(curvature,mincurv);
             uface=uface->next;
        }
        if(sides.first==sides.last){
            boundary[u]=1;
        }
        kill_list(&sides);
        
        return edgelength * curvature;
}

void ComputeEdgeCostAtVertex(MESH *m, int v) {
	/* compute the edge collapse cost for all edges that start
	   from vertex v.  Since we are only interested in reducing
	   the object by selecting the min cost edge at each step, we
	   only cache the cost of the least cost edge at this vertex
	   (collapse[v]) as well as the value of the
	   cost (objdist[v]). */
        INDEX_LIST_ENTRY *vneighbor;
        float dist;
        
	if(neighbor[v].length==0) {
		/* v doesn't have neighbors so it costs nothing to collapse */
		collapse[v]=-1;
		objdist[v]=-0.01f;
		return;
	}
	objdist[v] = 1000000;
	collapse[v]=-1;
	/* search all neighboring edges for "least cost" edge */
        vneighbor=neighbor[v].first;
        while(vneighbor!=NULL){
           dist = ComputeEdgeCollapseCost(m,v,vneighbor->index);

           if(dist<objdist[v]) {
			collapse[v]=vneighbor->index;  // candidate for edge collapse
			objdist[v]=dist;             // cost of the collapse
           }


           vneighbor=vneighbor->next;
        }
}

void ComputeAllEdgeCollapseCosts(MESH *m) {
	/* For all the edges, compute the difference it would make
	   to the model if it was collapsed.  The least of these
	   per vertex is cached in each vertex object.*/
        int i;
	for(i=0;i<m->num_vertices;i++) {
		ComputeEdgeCostAtVertex(m,i);

	}
}

int MinimumCostEdge(MESH *m){
	/* Find the edge that when collapsed will affect model the least. */
	int i,mn=0;
	for(i=0;i<m->num_vertices;i++) {
		if(objdist[i] < objdist[mn]) {
			mn = i;
		}
	}
	return mn;
}


int ProgressiveMesh(MESH *m)
{
        int mn;
        int i;
        if(!fill_face(m))
            return -1;
        if(!fill_neighbor(m))
            return -1;

        /****/
        boundary=malloc(sizeof(float)*m->num_vertices);
        if(boundary==NULL)
            return -1;
        for(i=0;i<m->num_vertices;i++)
            boundary[i]=0;
        /****/

        TNormals=malloc(sizeof(VECTOR)*m->num_faces);
        collapse=malloc(sizeof(int)*m->num_vertices);
        objdist=malloc(sizeof(float)*m->num_vertices);
        if(TNormals==NULL||collapse==NULL||objdist==NULL)
            return -1;
        get_triangle_normals(m->faces,TNormals,m->num_faces);
	ComputeAllEdgeCollapseCosts(m); // cache all edge collapse costs

        mn=MinimumCostEdge(m);

        if(boundary[mn] && !boundary[collapse[mn]]){
           meld_vertices(m, mn,collapse[mn]);
        }
        else
           meld_vertices(m, collapse[mn],mn);
        free(TNormals); TNormals=NULL;
        free(objdist); objdist=NULL;
        free(collapse); collapse=NULL;
        free(boundary);boundary=NULL;
        destroy_face(m);
        destroy_neighbor(m);
        
        return 1;
}

